@@ -23,7 +23,7 @@ case "${1:-default}" in
 
 	prerelease)
 		script authors
-		build transifex
+		build weblate
 		pushd man ; ./refresh.sh ; popd
 		git add -A gui man AUTHORS
 		git commit -m 'gui, man, authors: Update docs, translations, and contributors'
