@@ -17,6 +17,8 @@ import (
 	"sort"
 	"testing"
 	"time"
+
+	"github.com/syncthing/syncthing/lib/build"
 )
 
 func TestFakeFS(t *testing.T) {
@@ -119,7 +121,7 @@ func TestFakeFS(t *testing.T) {
 	}
 
 	// Chown
-	if err := fs.Lchown("dira", 1234, 5678); err != nil {
+	if err := fs.Lchown("dira", "1234", "5678"); err != nil {
 		t.Fatal(err)
 	}
 	if info, err := fs.Lstat("dira"); err != nil {
@@ -204,7 +206,6 @@ func TestFakeFSCaseSensitive(t *testing.T) {
 	}
 
 	testDir, sensitive := createTestDir(t)
-	defer removeTestDir(t, testDir)
 	if sensitive {
 		filesystems = append(filesystems, testFS{runtime.GOOS, newBasicFilesystem(testDir)})
 	}
@@ -240,7 +241,6 @@ func TestFakeFSCaseInsensitive(t *testing.T) {
 	}
 
 	testDir, sensitive := createTestDir(t)
-	defer removeTestDir(t, testDir)
 	if !sensitive {
 		filesystems = append(filesystems, testFS{runtime.GOOS, newBasicFilesystem(testDir)})
 	}
@@ -251,10 +251,7 @@ func TestFakeFSCaseInsensitive(t *testing.T) {
 func createTestDir(t *testing.T) (string, bool) {
 	t.Helper()
 
-	testDir, err := os.MkdirTemp("", "")
-	if err != nil {
-		t.Fatalf("could not create temporary dir for testing: %s", err)
-	}
+	testDir := t.TempDir()
 
 	if fd, err := os.Create(filepath.Join(testDir, ".stfolder")); err != nil {
 		t.Fatalf("could not create .stfolder: %s", err)
@@ -273,14 +270,6 @@ func createTestDir(t *testing.T) (string, bool) {
 	return testDir, sensitive
 }
 
-func removeTestDir(t *testing.T, testDir string) {
-	t.Helper()
-
-	if err := os.RemoveAll(testDir); err != nil {
-		t.Fatalf("could not remove test directory: %s", err)
-	}
-}
-
 func runTests(t *testing.T, tests []test, filesystems []testFS) {
 	for _, filesystem := range filesystems {
 		for _, test := range tests {
@@ -576,7 +565,7 @@ func testFakeFSRenameInsensitive(t *testing.T, fs Filesystem) {
 	}
 
 	// not checking on darwin due to https://github.com/golang/go/issues/35222
-	if runtime.GOOS != "darwin" {
+	if !build.IsDarwin {
 		if err := fs.Rename("/foo/bar/BAZ", "/FOO/BAR/bAz"); err != nil {
 			t.Errorf("Could not perform in-place case-only directory rename: %s", err)
 		}
@@ -799,7 +788,7 @@ func testFakeFSSameFile(t *testing.T, fs Filesystem) {
 			t.Fatalf("Could not create %s: %s", filename, err)
 		} else {
 			fd.Close()
-			if runtime.GOOS == "windows" {
+			if build.IsWindows {
 				time.Sleep(1 * time.Millisecond)
 			}
 		}
